-- Time-stamp: <2006-12-11 00:27:06 cklin>

module Infopipes.Compiler where

import Data.List
import Data.Maybe
import Infopipes.Syntax
import Infopipes.Polarity
import Infopipes.Utils

store = "__packet"
path_key = "__index"
path_cond index = concat ["    if (", path_key, " == ", index, ") "]

compile_stmt :: Port -> Statement -> Statement
compile_stmt negative_port = driver where
    driver stmt@(Single _)  = stmt
    driver (Block u v body) = Block u v (map driver body)
    driver (Cond branches)  = Cond (mapsnd driver branches)
        where mapsnd f = map (\(u, v) -> (u, f v))
    driver (Input var port rest) = Single $ concat $
        if (Inport, port) == negative_port
           then [var, " = ", store, rest]
           else [var, " = ", port_name port, ".pull()", rest]
    driver (Output port expr rest) = Single $ concat $
        if (Outport, port) == negative_port
           then [store, " = ", expr, rest]
           else [port_name port, ".push(", expr, ")", rest]

compile_path :: (Int, Port) -> Statement -> (Bool, Statement)
compile_path (index, port@(tag, name)) (Block u v body) =
    (tag == Inport, compile_stmt port indexed_path) where
        indexed_path = Cond [(path_cond (show index), body_return)]
        body_return  = Block u v (body ++ [Single return_stmt])
        return_stmt  = "  return " ++ return_value ++ "; "
        return_value = if tag == Inport then "" else store

compile :: Infopipe -> [PipeObject]
compile pipe = map translate (cartesian negv_choices) where
    lifted_paths = alt_lift_stmt (pipe_body pipe)
    poly_ports   = list_poly_ports lifted_paths
    all_ports    = pipe_ports pipe
    tag_index pt = (fromJust (elemIndex pt all_ports), pt)
    negv_choices = map (map tag_index) poly_ports
    translate negv = pipe_object where
        code_paths   = zipWith compile_path negv lifted_paths
        (push, pull) = partition fst code_paths
        polarity pt  = if elem pt (map snd negv) then 'N' else 'P'
        pipe_object  = PipeObject
            { push_paths = map snd push,
              pull_paths = map snd pull,
              posv_ports = zip [0..] all_ports \\ negv,
              obj_config = map polarity all_ports }

format_stmt :: Statement -> String
format_stmt = format where
    format (Single s)       = s
    format (Block u v body) = u ++ format_stmts body ++ v
    format (Cond branches)  = concatMap format_branch branches
    format_stmts = concatMap format_stmt
    format_branch (u, v) = u ++ format v

format_push :: [Statement] -> String
format_push paths = concat
    ["  public void push(int ", path_key,
     ", Object ", store, ") {\n",
     concatMap format_stmt paths,
     "    throw new RuntimeException();\n  }\n\n"]

format_pull :: [Statement] -> String
format_pull paths = concat
    ["  public Object pull(int ", path_key, ") {\n",
     "    Object ", store, ";\n",
     concatMap format_stmt paths,
     "    throw new RuntimeException();\n  }\n"]

format_link :: PortType -> (Int, Port) -> String
format_link target (index, (tag, name)) =
    if tag /= target then "" else concat
           ["  if (index == ", show index, ") {\n      ",
            port_name name, " = port;\n      return;\n    }\n  "]

format_connected :: Port -> String
format_connected (_, name) = concat
    ["    if (", port_name name, " == null)\n",
     "      throw new RuntimeException();\n"]

format_physical :: Infopipe -> PipeObject -> String
format_physical pipe pipe_obj = concat
    [pipe_package pipe, pipe_imports pipe,
     "import infopipes.pipes.*;\n",
     "import infopipes.utils.*;\n\n",
     "public class ", pipe_name pipe, "PhysicalPipe",
     obj_config pipe_obj, " extends ", pipe_super pipe,
     " implements PhysicalPipe {\n",
     "  Port ", commas $ map (port_name . snd) posv,
     ";\n\n  public Port getPush(int index) {\n",
     "    return new Port(this, index);\n",
     "  }\n\n",
     "  public Port getPull(int index) {\n",
     "    return new Port(this, index);\n",
     "  }\n\n",
     "  public void linkPush(int index, Port port) {\n  ",
     concatMap (format_link Outport) indexed_posv,
     "  throw new RuntimeException();\n  }\n\n",
     "  public void linkPull(int index, Port port) {\n  ",
     concatMap (format_link Inport) indexed_posv,
     "  throw new RuntimeException();\n  }\n\n",
     "  public void connected() {\n",
     concatMap format_connected posv,
     "  }\n\n",
     format_push (push_paths pipe_obj),
     format_pull (pull_paths pipe_obj), "}\n"]
        where indexed_posv = posv_ports pipe_obj
              posv = map snd indexed_posv

format_configs :: (Int, String) -> String
format_configs (index, config) = concat
    ["  static Polarity[] config", show index, " = { ",
     commas $ map (\c -> "Polarity." ++ [c]) config, " };\n"]

format_port_stmt :: (Int, Port) -> String
format_port_stmt (index, (tag, name)) = concat
    ["    ", if tag == Inport then "in" else "out",
     "ports.put(\"", port_name name, "\", ", show index, ");\n"]

format_config_stmt :: Int -> String
format_config_stmt index =
    "    configs.add(config" ++ show index ++ ");\n"

format_instantiate :: String -> (Int, String) -> String
format_instantiate name (index, config) = concat
    ["    if (config.equals(config", show index, "))\n",
     "      return new ", name, "PhysicalPipe", config, "();\n"]

format_logical :: Infopipe -> [String] -> String
format_logical pipe configs = concat
    [pipe_package pipe,
     "import infopipes.pipes.*;\n",
     "import infopipes.utils.*;\n\n",
     "public class ", name,
     "LogicalPipe extends BaseLogicalPipe {\n",
     concatMap format_configs indexed_configs,
     "\n  public ", name, "LogicalPipe() {\n",
     concatMap format_port_stmt (zip [0..] (pipe_ports pipe)),
     concatMap format_config_stmt (map fst indexed_configs),
     "  }\n\n  @Override\n  public PhysicalPipe ",
     "instantiate(PolarityConfig config) {\n",
     concatMap (format_instantiate name) indexed_configs,
     "    throw new RuntimeException();\n  }\n}\n"] where
        name = pipe_name pipe
        indexed_configs = zip [0..] configs

