-- Time-stamp: <2006-12-10 16:04:11 cklin>

module Infopipes.Parser where

import Data.List
import Text.ParserCombinators.Parsec
import Infopipes.Syntax
import Infopipes.Utils

type Lexer = Parser String

-- Make variants of standard parsec parsers so that they return a string
-- (instead of a character or a list of strings).  These variants
-- greatly simplify the use of combinators.

flat_lexer p = p >>= return . concat
list_lexer p = p >>= return . (:[])

multi    = flat_lexer . many
multi1   = flat_lexer . many1
cchar    = list_lexer . char
cnoneOf  = list_lexer . noneOf
canyChar = list_lexer anyChar
cspace   = list_lexer space <|> comment
cspaces  = multi cspace

-- The comment lexer recognizes Java-style single-line and non-nested
-- block comments.

comment = try line_comment <|> try block_comment where
    line_comment  = string "//" <+> line_body <+> line_end
    line_body     = multi (cnoneOf "\n")
    line_end      = try canyChar <|> return ""
    block_comment = string "/*" <+> in_comment
    block_end     = string "*/"
    in_comment    = block_end <|> (canyChar <+> in_comment)

-- The <+> operator sequences two lexers and returns their results
-- concatenated into a single string; it is useful for recording input
-- strings verbatim through syntax matching.

(<+>) :: Lexer -> Lexer -> Lexer
(<+>) first second =
    do x <- first
       y <- second
       return (x ++ y)

-- The sepWith1 combinator is just like sepBy1 except that it also
-- records the output of the separator lexers.

sepWith1 :: Lexer -> Lexer -> Lexer
sepWith1 meat sep = meat <+> multi (sep <+> meat)

-- The lexesc lexer matches the input string with backslash escape
-- until it encounters a character in the exclusion list.

lexesc :: [Char] -> Lexer
lexesc ex = quoted <|> unquoted where
    quoted   = cchar '\\' <+> canyChar
    unquoted = cnoneOf ex

-- The keyword lexer matches the given keyword string.  It differs from
-- the string lexer in that it would not match the prefix of a string.

keyword :: String -> Lexer
keyword word = try $
    do string word
       notFollowedBy alphaNum
       return word <+> cspaces

-- The symbol lexer matches the given character symbol and consumes any
-- subsequent spaces.

symbol :: Char -> Lexer
symbol sym = cchar sym <+> cspaces

-- The literal lexer matches a Java identifier, which consists of a
-- letter, a dollar sign, or an underscore, followed by zero or more
-- alpha-numerical, dollar sign, or underscore characters.  This lexer
-- does not recognize Unicode escapes or Unicode characters.

literal :: Lexer
literal =
    do first <- letter <|> oneOf "$_"
       rest <- many (alphaNum <|> oneOf "$_")
       cspaces
       return (first:rest)

-- The char_const and string_const lexers matches (single-quote)
-- character and (double-quote) string constants.  I do not really care
-- about the constants per se, but the lexer needs to know about them
-- because they introduce different syntax classes.

char_const :: Lexer
char_const = quote <+> content <+> quote where
    quote   = cchar '\''
    content = lexesc "'"

string_const :: Lexer
string_const = quote <+> content <+> quote where
    quote   = cchar '"'
    content = multi (lexesc "\"")

-- The parens lexers record a code segment enclosed in a pair of
-- parentheses.  The lexer needs to know about parenthesis in order to
-- recognize for and if constructs correctly.

parens :: Lexer
parens = lquote <+> code_segment <+> rquote where
    lquote = symbol '('
    rquote = symbol ')'

parens1 :: Lexer
parens1 = lquote <+> code_segment1 <+> rquote where
    lquote = symbol '('
    rquote = symbol ')'

-- The code_segment lexer recognizes self-contained text strings
-- (including spaces) that are insignificant for parsing purposes.

code_entity :: Lexer
code_entity = choice lexers where
    lexers  = [litchar, parens, char_const, string_const]
    litchar = cnoneOf "{}();'\""

code_segment :: Lexer
code_segment = multi code_entity

code_segment1 :: Lexer
code_segment1 = multi1 code_entity

-- The single_stmt, block_stmt, for_stmt, and if_stmt parsers recognize
-- different types of Java statements.  The single_stmt parser is quite
-- lax about Java syntax matching and will accept lots of ill-formed
-- Java statements without complaint.

single_stmt :: Parser Statement
single_stmt =
    do stmt <- code_segment <+> symbol ';'
       return (Single stmt)

block_stmt :: Parser Statement
block_stmt =
    do lquote <- symbol '{'
       body <- many statement
       rquote <- symbol '}'
       return (Block lquote rquote body)

-- I tag all loop and conditional statements with the Cond data
-- constructor because their only effect is to prevent partially-
-- accessed ports from acquiring a negative polarity.

null_stmt = Single ""
null_branch = ("", null_stmt)

for_stmt :: Parser Statement
for_stmt =
    do lquote <- keyword "for" <+> symbol '('
       condition <- sepWith1 code_segment (cchar ';')
       rquote <- symbol ')'
       body <- statement
       let prefix = concat [lquote, condition, rquote]
       return (Cond [(prefix, body), null_branch])

do_while_stmt :: Parser Statement
do_while_stmt =
    do prefix <- keyword "do"
       body <- statement
       suffix <- keyword "while" <+> parens1 <+> symbol ';'
       return (Cond [(prefix, body), (suffix, null_stmt)])

while_stmt :: Parser Statement
while_stmt =
    do prefix <- keyword "while" <+> parens1
       body <- statement
       return (Cond [(prefix, body), null_branch])

if_stmt :: Parser Statement
if_stmt =
    do prefix <- keyword "if" <+> parens1
       true_branch <- statement
       let if_part = (prefix, true_branch)
       else_part <- option [null_branch]
           (do middle <- keyword "else"
               false_branch <- statement
               return [(middle, false_branch)])
       return (Cond (if_part:else_part))

-- The input_stmt and output_stmt parsers match DirectFlow input (?) and
-- output (!) statements.  Need to enclose these computations within a
-- try construct because neither statement starts with a key symbol.

input_stmt :: Parser Statement
input_stmt = try $
    do store <- literal
       symbol '='
       port <- literal
       symbol '?'
       rest <- symbol ';'
       return (Input store (PortName port) rest)

output_stmt :: Parser Statement
output_stmt = try $
    do port <- literal
       symbol '!'
       value <- code_segment1
       rest <- symbol ';'
       return (Output (PortName port) value rest)

alt_stmt :: Parser Statement
alt_stmt =
    do keyword "alt"
       branches <- statement `sepBy1` keyword "with"
       return (Alt branches)

statement :: Parser Statement
statement = choice parsers where
    parsers = [input_stmt, output_stmt, alt_stmt, if_stmt, for_stmt,
               do_while_stmt, while_stmt, single_stmt, block_stmt]

-- The declaration parser matches a regular C/Java style variable
-- declaration statement: it matches the given keyword and returns a
-- list of declared variables (ports).

declaration :: (String -> a) -> String -> Parser [a]
declaration tag key =
    do keyword key
       ports <- literal `sepBy1` symbol ','
       symbol ';'
       return (map tag ports)

inport_spec :: Parser [Port]
inport_spec = declaration ((,) Inport . PortName) "inport"

outport_spec :: Parser [Port]
outport_spec = declaration ((,) Outport . PortName) "outport"

port_spec :: Parser [Port]
port_spec = inport_spec <|> outport_spec

-- A DirectFlow program is an Infopipe definition prefixed by a
-- mandatory Java package declaration.

module_decl :: String -> Lexer
module_decl key =
    keyword key <+>
    sepWith1 (cchar '*' <|> literal) (cchar '.') <+>
    cspaces <+> symbol ';'

program :: Parser Infopipe
program =
    do package <- module_decl "package"
       imports <- multi (module_decl "import")
       keyword "pipe"
       name <- literal
       let base = do keyword "extends" ; literal
       super <- option "Object" base
       symbol '{'
       ports <- multi1 port_spec
       keyword "process"
       body <- block_stmt
       symbol '}'
       return $ Infopipe
                  { pipe_package = package,
                    pipe_imports = imports,
                    pipe_super = super,
                    pipe_name = name,
                    pipe_ports = ports,
                    pipe_body = body }

