-- Time-stamp: <2006-12-05 15:58:31 cklin>

module Infopipes.Tests.Parser where

import Test.HUnit
import Text.ParserCombinators.Parsec
import Infopipes.Syntax
import Infopipes.Parser

--

error_col :: ParseError -> Int
error_col = sourceColumn . errorPos

unexpected_failure :: ParseError -> String
unexpected_failure err = template ++ location where
    template = " unexpected parse failure at column "
    location = show $ error_col err

unexpected_success :: Show a => a -> String
unexpected_success result = template ++ show result where
    template = " unexpected parse of "

--

good_parse :: (Eq a, Show a) => Parser a -> a -> String -> Test
good_parse parser answer input = TestCase $
    case parse parser "" input of
      Left err -> assertFailure (unexpected_failure err)
      Right result -> assertEqual prefix answer result
    where prefix = " examining parse result"

bad_parse :: Show a => Parser a  -> Column -> String-> Test
bad_parse parser col input = TestCase $
    case parse parser "" input of
      Left err -> assertEqual prefix col (error_col err)
      Right result -> assertFailure (unexpected_success result)
    where prefix = " examining parse error position"

-- 

cspaces_tests = TestLabel "cspaces" $ TestList
    [cspaces_0, cspaces_1, cspaces_2, cspaces_3, cspaces_4,
     cspaces_5, cspaces_6, cspaces_7, cspaces_8]

cspaces_0 = good_parse cspaces "" ""
cspaces_1 = good_parse cspaces "" "a"
cspaces_2 = good_parse cspaces " \t\n" " \t\n"
cspaces_3 = good_parse cspaces " \t\n" " \t\n3"
cspaces_4 = good_parse cspaces "// aab" "// aab"
cspaces_5 = good_parse cspaces " // aab\n " " // aab\n k"
cspaces_6 = good_parse cspaces "/* aab */ " "/* aab */ 7\n "
cspaces_7 = good_parse cspaces "/* // */ " "/* // */ 7\n "
cspaces_8 = good_parse cspaces "// /* aaa \n " "// /* aaa \n */"

--

keyword_tests = TestLabel "keyword" $ TestList
    [keyword_0, keyword_1, keyword_2]

keyword_0 = good_parse (keyword "for") "for" "for"
keyword_1 = good_parse (keyword "for") "for " "for ("
keyword_2 = bad_parse  (keyword "for") 1 "forall "

--

literal_tests = TestLabel "literal" $ TestList
    [literal_0, literal_1, literal_2, literal_3, literal_4,
     literal_5, literal_6, literal_7, literal_8]

literal_0 = good_parse literal "foo12" "foo12"
literal_1 = good_parse literal "foo12" "foo12 a3.5"
literal_2 = good_parse literal "foo12" "foo12.a3.5"
literal_3 = good_parse literal "foo12_3" "foo12_3 bar"
literal_4 = good_parse literal "$" "$"
literal_5 = good_parse literal "$35" "$35"
literal_6 = good_parse literal "_a$2" "_a$2"
literal_7 = bad_parse  literal 1 "12foo"
literal_8 = bad_parse  literal 1 ""

--

char_const_tests = TestLabel "char_const" $ TestList
    [char_const_0, char_const_1, char_const_2, char_const_3,
     char_const_4, char_const_5, char_const_6]

good_char_const = good_parse char_const
bad_char_const = bad_parse char_const

char_const_0 = good_char_const "'a'" "'a' "
char_const_1 = good_char_const "'\\b'" "'\\b' "
char_const_2 = good_char_const "'\\\\'" "'\\\\' "
char_const_3 = good_char_const "'\\''" "'\\'' "
char_const_4 = bad_char_const 3 "'ab'"
char_const_5 = bad_char_const 4 "'\\'"
char_const_6 = bad_char_const 2 "'' "

--

string_const_tests = TestLabel "string_const" $ TestList
    [string_const_0, string_const_1, string_const_2,
     string_const_3, string_const_4]

good_string_const = good_parse string_const
bad_string_const = bad_parse string_const

string_const_0 = good_string_const "\"\"" "\"\""
string_const_1 = good_string_const "\"a\"" "\"a\" "
string_const_2 = good_string_const "\"a'b\"" "\"a'b\" "
string_const_3 = good_string_const "\"a\\\"\"" "\"a\\\"\" "
string_const_4 = bad_string_const 4 "\"ab"

-- 

code_segment_tests = TestLabel "code_segment" $ TestList
    [code_segment_0, code_segment_1, code_segment_2,
     code_segment_3, code_segment_4, code_segment_5,
     code_segment_6, code_segment1_0, code_segment1_1,
     code_segment1_2, code_segment1_3, code_segment1_4,
     code_segment1_5, code_segment1_6]

good_code_segment = good_parse code_segment
bad_code_segment = bad_parse code_segment

code_segment_0 = good_code_segment "ab(4.5e3, 2)" "ab(4.5e3, 2)"
code_segment_1 = good_code_segment "ab()  " "ab()  "
code_segment_2 = good_code_segment "c.d(1+x).e" "c.d(1+x).e; "
code_segment_3 = good_code_segment "ab " "ab ) "
code_segment_4 = good_code_segment "" ""
code_segment_5 = bad_code_segment 8 "(a64.3 })"
code_segment_6 = bad_code_segment 8 "(a64.3 ; xy)"

good_code_segment1 = good_parse code_segment1
bad_code_segment1 = bad_parse code_segment1

code_segment1_0 = good_code_segment1 "ab(4.5e3, 2)" "ab(4.5e3, 2)"
code_segment1_1 = good_code_segment1 "ab()  " "ab()  "
code_segment1_2 = good_code_segment1 "c.d(1+x).e" "c.d(1+x).e; "
code_segment1_3 = good_code_segment1 "ab " "ab ) "
code_segment1_4 = bad_code_segment1 1 ""
code_segment1_5 = bad_code_segment1 8 "(a64.3 })"
code_segment1_6 = bad_code_segment1 8 "(a64.3 ; xy)"

--

stmt_tests = TestLabel "statement" $ TestList
    [block_stmt_tests, for_stmt_tests, if_stmt_tests,
     input_stmt_tests, output_stmt_tests, alt_stmt_tests]

good_stmt = good_parse statement
bad_stmt = bad_parse statement

--

block_stmt_tests = TestLabel "block_stmt" $ TestList
    [block_stmt_0, block_stmt_1, block_stmt_2,
     block_stmt_3, block_stmt_4, block_stmt_5,
     block_stmt_6]

block_stmt_0 = good_stmt answer "{ }" where
    answer = Block "{ " "}" []
block_stmt_1 = good_stmt answer "{ a(); }" where
    answer = Block "{ " "}" [Single "a(); "]
block_stmt_2 = good_stmt answer "{ a(); b; }" where
    answer = Block "{ " "}" [Single "a(); ", Single "b; "]
block_stmt_3 = good_stmt answer "{ a();; }" where
    answer = Block "{ " "}" [Single "a();", Single "; "]
block_stmt_4 = good_stmt answer "{ {;} }" where
    answer = Block "{ " "}" [Block "{" "} " [Single ";"]]
block_stmt_5 = bad_stmt 7 "{ a() }"
block_stmt_6 = bad_stmt 5 "{ a(; }"

--

for_stmt_tests = TestLabel "for_stmt" $ TestList
    [for_stmt_0, for_stmt_1, for_stmt_2, for_stmt_3,
     for_stmt_4, for_stmt_5]

for_stmt_0 = good_stmt answer "for ( ; ; ) ;" where
    answer = Cond [("for ( ; ; ) ", Single ";"), null_branch]
for_stmt_1 = good_stmt answer "for () ;" where
    answer = Cond [("for () ", Single ";"), null_branch]
for_stmt_2 = good_stmt answer "for(;) foo();" where
    answer = Cond [("for(;) ", Single "foo();"), null_branch]
for_stmt_3 = good_stmt answer "for (;;) foo(); bar();" where
    answer = Cond [("for (;;) ", Single "foo(); "), null_branch]
for_stmt_4 = bad_stmt 7 "for(,)"
for_stmt_5 = bad_stmt 5 "for({)"
for_stmt_6 = bad_stmt 1 "forall()"

--

if_stmt_tests = TestLabel "if_stmt" $ TestList
    [if_stmt_0, if_stmt_1, if_stmt_2, if_stmt_3, if_stmt_4]

if_stmt_0 = good_stmt answer "if (a) ;" where
    answer = Cond [("if (a) ", Single ";"), null_branch]
if_stmt_1 = good_stmt answer "if(a)b; else c;" where
    answer = Cond [("if(a)", Single "b; "), ("else ", Single "c;")]
if_stmt_2 = good_stmt answer "if(a)if(b)c;" where
    answer = Cond [("if(a)", Cond [("if(b)", Single "c;"),
                   null_branch]), null_branch]
if_stmt_3 = bad_stmt 4 "if();"
if_stmt_4 = bad_stmt 6 "if( a;)b;"

--

inport var port = Input var (PortName port)
outport port expr = Output (PortName port) expr

--

input_stmt_tests = TestLabel "input_stmt" $ TestList
    [input_stmt_0, input_stmt_1, input_stmt_2, input_stmt_3]

input_stmt_0 = good_stmt (inport "a" "b" ";") "a = b?;"
input_stmt_1 = good_stmt (Single "a = b.c()?;") "a = b.c()?;"
input_stmt_2 = good_stmt (Single "a() = b?;") "a() = b?;"
input_stmt_3 = bad_stmt 11 "a = b.c()?"

--

output_stmt_tests = TestLabel "output_stmt" $ TestList
    [output_stmt_0, output_stmt_1, output_stmt_2]

output_stmt_0 = good_stmt (outport "a" "b" ";") "a ! b;"
output_stmt_1 = good_stmt (outport "a" "b.c() " ";") "a ! b.c() ;"
output_stmt_2 = good_stmt (Single "a.b ! c;") "a.b ! c;"
output_stmt_3 = bad_stmt 6 "a ! b"

--

alt_stmt_tests = TestLabel "alt_stmt" $ TestList
    [alt_stmt_0, alt_stmt_1, alt_stmt_2]

alt_stmt_0 = good_stmt (Alt [Single "; "]) "alt ; "
alt_stmt_1 = good_stmt answer "alt a; with b;" where
    answer = Alt [Single "a; ", Single "b;"]
alt_stmt_2 = good_stmt (Alt [Single "a; "]) "alt a; b;"

--

unit_tests = runTestTT $ TestList
    [cspaces_tests, keyword_tests, literal_tests,
     char_const_tests, string_const_tests, code_segment_tests,
     stmt_tests]

